"""
A class for constructing Entity classes from EntityModels
"""

from __future__ import annotations

import builtins
from pathlib import Path
from typing import Annotated, Any, Dict, List, Literal, Tuple, Type

from pydantic import Field, create_model
from pydantic_core import PydanticUndefined
from ruamel.yaml.main import YAML

from ibek.globals import JINJA

from .ioc import Entity, EnumVal, clear_entity_model_ids
from .parameters import EnumParam, IdParam, ObjectParam
from .support import EntityModel, Support
from .utils import UTILS


class EntityFactory:
    def __init__(self) -> None:
        """
        A class to create `Entity` models from `EntityModel`s.

        Created models are stored in `self._entity_models` to lookup when
        resolving nested `SubEntity`s.
        """
        self._entity_models: Dict[str, Type[Entity]] = {}
        # starting a new EntityFactory implies we should throw away any existing
        # Entity instances - this is required for tests which create multiple
        # EntityFactories
        clear_entity_model_ids()

    def make_entity_models(self, entity_model_yaml: List[Path]) -> List[Type[Entity]]:
        """
        Read a set of *.ibek.support.yaml files and generate Entity classes
        from their EntityModel entries
        """

        try:
            for entity_model in entity_model_yaml:
                support_dict = YAML(typ="safe").load(entity_model)

                Support.model_validate(support_dict)

                # deserialize the support module yaml file
                support = Support(**support_dict)
                # make Entity classes described in the support module yaml file
                self._make_entity_models(support)
        except Exception:
            print(f"VALIDATION ERROR READING {entity_model}")
            raise

        return list(self._entity_models.values())

    def _make_entity_model(self, model: EntityModel, support: Support) -> Type[Entity]:
        """
        Create an Entity Model from a EntityModel instance and a Support instance.
        """

        def add_arg(name, typ, description, default):
            if default is None:
                default = PydanticUndefined
            # cheesy check for enum, can this be improved?
            if typ in [str, object] or "enum" in str(typ):
                args[name] = (Annotated[typ, Field(description=description)], default)
            else:
                args[name] = (
                    Annotated[
                        Annotated[
                            str,
                            Field(
                                description=f"jinja that renders to {typ}",
                                pattern=JINJA,
                            ),
                        ]
                        | Annotated[typ, Field(description=description)],
                        Field(
                            description=f"union of {typ} and jinja "
                            "representation of {typ}"
                        ),
                    ],
                    default,
                )

        args: Dict[str, Tuple[type, Any]] = {}
        validators: Dict[str, Any] = {}

        # fully qualified name of the Entity class including support module
        full_name = f"{support.module}.{model.name}"

        # add in each of the parameters as a Field in the Entity
        for name, param in model.parameters.items():
            type: Any

            if isinstance(param, ObjectParam):
                # we now defer the lookup of the object until whole model validation
                type = object

            elif isinstance(param, IdParam):
                type = str

            elif isinstance(param, EnumParam):
                # Pydantic uses the values of the Enum as the options in the schema.
                # Here we arrange for the keys to be in the schema (what a user supplies)
                # but the values to be what is rendered when jinja refers to the enum
                enum_swapped = {}
                for k, v in param.values.items():
                    enum_swapped[str(v) if v else str(k)] = k
                # TODO review enums especially with respect to Pydantic 2.7.1
                val_enum = EnumVal(name, enum_swapped)  # type: ignore
                type = val_enum

            else:
                # arg.type is str, int, float, etc.
                type = getattr(builtins, param.type)
            add_arg(name, type, param.description, getattr(param, "default"))  # type: ignore

        # add the type literal which discriminates between the different Entity classes
        typ = Literal[full_name]  # type: ignore
        args["type"] = (typ, Field(description=model.description))

        class_name = full_name.replace(".", "_")
        entity_cls = create_model(
            class_name,
            **args,
            __validators__=validators,
            __base__=Entity,
        )  # type: ignore

        # add a link back to the EntityModel Instance that generated this Entity Class
        entity_cls._model = model

        # store this Entity class in the factory
        self._entity_models[full_name] = entity_cls

        return entity_cls

    def _make_entity_models(self, support: Support) -> List[Type[Entity]]:
        """
        Create Entity subclasses for all EntityModel instances in the given
        Support instance. Returns a list of the Entity subclasses Models.
        """
        entity_names = []
        entity_models = []

        for model in support.entity_models:
            if model.name in entity_names:
                # not tested because schema validation will always catch this first
                raise ValueError(f"Duplicate entity name {model.name}")

            entity_models.append(self._make_entity_model(model, support))

            entity_names.append(model.name)
        return entity_models

    def resolve_sub_entities(self, entities: List[Entity]) -> List[Entity]:
        """
        Recursively resolve SubEntity collections in a list of Entity instances
        """
        resolved_entities: List[Entity] = []
        for parent_entity in entities:
            model = parent_entity._model
            # add the parent standard entity
            resolved_entities.append(parent_entity)
            # add in SubEntities if any
            for sub_entity in model.sub_entities:
                # find the Entity Class that the SubEntity represents
                entity_cls = self._entity_models[sub_entity.type]
                # get the SubEntity arguments
                sub_params_dict = sub_entity.model_dump()
                # jinja render any references to parent Params in the SubEntity Args
                for key, param in sub_params_dict.items():
                    sub_params_dict[key] = UTILS.render(parent_entity, param)
                # cast the SubEntity to its concrete Entity subclass
                entity = entity_cls(**sub_params_dict)
                # recursively scan the SubEntity for more SubEntities
                resolved_entities.extend(self.resolve_sub_entities([entity]))
        return resolved_entities



"""
The EntityModel class describes what a given support module can instantiate.
"""

from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Mapping, Optional, Sequence, Union

from pydantic import Field, PydanticUndefinedAnnotation
from typing_extensions import Literal

from .globals import BaseSettings
from .parameters import Define, IdParam, Param
from .sub_entity import SubEntity


def default(T: type):
    """
    defines a default type which may be
    """
    return Field(
        Union[Optional[T], PydanticUndefinedAnnotation],
        description="If given, instance doesn't supply argument, what value should be used",
    )


class When(Enum):
    first = "first"
    every = "every"
    last = "last"


class Database(BaseSettings):
    """
    A database file that should be loaded by the startup script and its args
    """

    file: str = Field(
        description="Filename of the database template in <module_root>/db"
    )

    enabled: str = Field(
        description="Set to False to disable loading this database", default="True"
    )

    args: Mapping[str, Optional[str]] = Field(
        description=(
            "Dictionary of args and values to pass through to database. "
            "A value of None is equivalent to ARG: '{{ ARG }}'. "
            "See `UTILS.render_map` for more details."
        )
    )


class EnvironmentVariable(BaseSettings):
    """
    An environment variable that should be set in the startup script
    """

    name: str = Field(description="Name of environment variable")
    value: str = Field(description="Value to set")


class Comment(BaseSettings):
    """
    A script snippet that will have '# ' prepended to every line
    for insertion into the startup script
    """

    type: Literal["comment"] = "comment"
    when: When = Field(description="One of first / every / last", default="every")
    value: str = Field(
        description="A comment to add into the startup script", default=""
    )


class Text(BaseSettings):
    """
    A script snippet to insert into the startup script
    """

    type: Literal["text"] = "text"
    when: str = Field(description="One of first / every / last", default="every")
    value: str = Field(description="raw text to add to the startup script", default="")


Script = Sequence[Union[Text, Comment]]


class EntityPVI(BaseSettings):
    """Entity PVI definition"""

    yaml_path: str = Field(
        description="Path to .pvi.device.yaml - absolute or relative to PVI_DEFS"
    )
    ui_index: bool = Field(
        True,
        description="Whether to add the UI to the IOC index.",
    )
    ui_macros: dict[str, str | None] = Field(
        None,
        description=(
            "Macros to launch the UI on the IOC index. "
            "These must be args of the Entity this is attached to."
        ),
    )
    pv: bool = Field(
        False,
        description=(
            "Whether to generate a PVI PV. This adds a database template with info "
            "tags that create a PVAccess PV representing the device structure."
        ),
    )
    pv_prefix: str = Field("", description='PV prefix for PVI PV - e.g. "$(P)"')


discriminated = Annotated[  # type: ignore
    Union[tuple(Param.__subclasses__())],
    Field(discriminator="type", description="union of arg types"),
]


class EntityModel(BaseSettings):
    """
    A Model for a class of Entity that an IOC instance may instantiate
    """

    name: str = Field(
        description="Publish EntityModel as type <module>.<name> for IOC instances"
    )
    description: str = Field(
        description="A description of the Support module defined here"
    )
    pre_defines: dict[str, Define] = Field(
        description="Calculated values to use as additional arguments "
        "With Jinja evaluation before all Args",
        default=(),
    )
    # declare Arg as Union of its subclasses for Pydantic to be able to deserialize
    parameters: dict[str, discriminated] = Field(  # type: ignore
        description="The arguments IOC instance should supply",
        default=(),
    )
    post_defines: dict[str, Define] = Field(
        description="Calculated values to use as additional arguments "
        "With Jinja evaluation after all Args",
        default=(),
    )
    pre_init: Script = Field(
        description="Startup script snippets to add before iocInit()", default=()
    )
    post_init: Script = Field(
        description="Startup script snippets to add post iocInit(), such as dbpf",
        default=(),
    )
    databases: Sequence[Database] = Field(
        description="Databases to instantiate", default=[]
    )
    env_vars: Sequence[EnvironmentVariable] = Field(
        description="Environment variables to set in the boot script", default=()
    )
    pvi: Optional[EntityPVI] = Field(
        description="PVI definition for Entity", default=None
    )

    # list of additional entities to instantiate for each instance of this definition
    sub_entities: Sequence[SubEntity] = Field(
        description="The sub-entity instances that this collection is to instantiate",
        default=(),
    )

    shared: Sequence[Any] = Field(
        description="A place to create any anchors required for repeating YAML",
        default=(),
    )

    def _get_id_arg(self) -> str | None:
        """
        Lookup which of this object's fields represents its ID. The ID is the
        field of type IdParam whose value is the identifying name for
        this object.

        The jinja templates of other objects can refer to this object by the
        value of its ID field.
        """
        for name, param in self.parameters.items():
            if isinstance(param, IdParam):
                return name
        return None



"""
some system tests
"""

import os
import re
import shutil
import subprocess
from pathlib import Path

import pytest


def run_command(command: str, error_OK=False, show=False):
    """
    Run a command and print the output to the console. Throw an exception on failure.
    """
    print(f"Running: {command}")

    p_result = subprocess.run(command, capture_output=True, shell=True)

    output = p_result.stdout.decode()
    error_out = p_result.stderr.decode()

    result = output + error_out

    if p_result.returncode != 0:
        msg = f"Command Failed: {command}\n\n{result}\n\n\n"
        raise RuntimeError(msg)

    print(result)


@pytest.mark.skipif(
    os.getenv("IBEK_SYSTEM_TESTING") != "true",
    reason="export IBEK_SYSTEM_TESTING=true",
)
@pytest.mark.skipif(
    os.getenv("REMOTE_CONTAINERS") == "true", reason="only run outside devcontainers"
)
def test_container_build_and_run(tmp_path: Path):
    """
    make sure that a container build works and that the container can run an IOC

    IMPORTANT: this test runs against the ibek-test-KEEP branch of the
    ioc-template-example project.

    To update - clone ioc-template-example repo, then:
      - 'copier update --trust'.
      - update the ibek-support submodule to the desired version.
      - update its requirements.txt to point a compatible (this) version of ibek
      - push the changes to ibek-test-KEEP branch.
      - verify that it's CI tests pass.
    """
    ioc = "ioc-template-example"

    # get the ioc-template-example generic container source
    os.chdir(tmp_path)
    run_command(f"git clone https://github.com/epics-containers/{ioc}")
    run_command(f"cd {ioc} && git checkout ibek-test-KEEP")

    # patch the dockerfile to include this version of ibek
    ibek = Path(__file__).parent.parent
    shutil.copytree(ibek, tmp_path / ioc / "ibek")
    docker = tmp_path / ioc / "Dockerfile"
    text = docker.read_text()
    this_ibek = "COPY ibek ibek\nRUN pip install ./ibek"
    text = re.sub(r"RUN pip install.*\n", this_ibek, text)
    docker.write_text(text)
    print(text)

    # build the container with latest ibek and latest ibek-support
    os.chdir(ioc)
    run_command("git submodule update --init")
    # use the submodule version that comes with ibek-test-KEEP branch
    # so don't do the checkout below.
    # run_command("cd ibek-support && git checkout main")
    run_command("./build")

    # run the container test script
    run_command("./tests/run-tests.sh")



"""
A few global definitions
"""

import os
from pathlib import Path

from pydantic import BaseModel, ConfigDict
from typer.core import TyperGroup

DEFAULT_ARCH = "linux-x86_64"


class _Globals:
    """
    Helper class for accessing global constants.

    These constants define the paths to the various directories used by the
    ibek commands.
    """

    def __init__(self) -> None:
        """Initialize the global constants."""

        # Can be overridden by defining an environment variable "EPICS_ROOT"
        self._EPICS_ROOT = Path(os.getenv("EPICS_ROOT", "/epics/"))

        self._DEFAULT_ARCH = "linux-x86_64"

    @property
    def EPICS_ROOT(self):
        """Root of epics directory tree"""
        return self._EPICS_ROOT

    @property
    def SUPPORT(self):
        """Directory containing support module clones"""
        return self._EPICS_ROOT / "support"

    @property
    def RELEASE(self):
        """The global RELEASE file which lists all support modules"""
        return self._EPICS_ROOT / "support" / "configure" / "RELEASE"

    @property
    def RUNTIME_OUTPUT(self):
        """Directory containing runtime generated assets for IOC boot."""
        return self._EPICS_ROOT / "runtime"

    @property
    def EPICS_TARGET_ARCH(self):
        """The target architecture for the current container."""
        return os.getenv("EPICS_TARGET_ARCH", self._DEFAULT_ARCH)

    @property
    def EPICS_HOST_ARCH(self):
        """The host architecture for the current container."""
        return os.getenv("EPICS_HOST_ARCH", self._DEFAULT_ARCH)

    @property
    def NATIVE(self):
        """True if the target architecture is the same as the host architecture."""
        return self.EPICS_TARGET_ARCH == self.EPICS_HOST_ARCH

    @property
    def STATIC_BUILD(self):
        """True if the target architecture is not the default architecture."""
        return os.getenv("STATIC_BUILD", self.EPICS_TARGET_ARCH != self._DEFAULT_ARCH)

    @property
    def IBEK_DEFS(self):
        """Directory containing ibek support yaml files."""
        return self._EPICS_ROOT / "ibek-defs"

    @property
    def PVI_DEFS(self):
        """Directory containing pvi device yaml files."""
        return self._EPICS_ROOT / "pvi-defs"

    @property
    def OPI_OUTPUT(self):
        """Directory containing runtime generated opis to serve over http."""
        return self._EPICS_ROOT / "opi"

    @property
    def EPICS_BASE(self):
        """The folder containing the epics base source and binaries"""
        return self._EPICS_ROOT / "epics-base"

    @property
    def IOC_FOLDER(self):
        """root folder of a generic IOC source inside the container"""
        return self._EPICS_ROOT / "ioc"

    @property
    def CONFIG_DIR_NAME(self):
        """Name of config directory within IOC directory"""
        return "config"

    @property
    def IOC_DIR_NAME(self):
        """folder of the IOC source"""
        return "ioc"

    @property
    def RELEASE_SH(self):
        """a bash script to export the macros defined in RELEASE as environment vars"""
        return self.SUPPORT / "configure" / "RELEASE.shell"

    @property
    def MODULES(self):
        """global MODULES file used to determine order of build"""
        return self.SUPPORT / "configure" / "MODULES"

    @property
    def IOC_DBDS(self):
        """ibek-support list of declared dbds"""
        return self.SUPPORT / "configure" / "dbd_list"

    @property
    def IOC_LIBS(self):
        """ibek-support list of declared libs"""
        return self.SUPPORT / "configure" / "lib_list"

    @property
    def RUNTIME_DEBS(self):
        """ibek-support list of declared deb packages to install in runtime stage"""
        return self.SUPPORT / "configure" / "runtime_debs"

    @property
    def RUNTIME_FILES(self):
        """ibek-support list of files to copy to the runtime stage"""
        return self.SUPPORT / "configure" / "runtime_files_list"


# Folder containing templates for IOC src etc.
TEMPLATES = Path(__file__).parent / "templates"

# Path suffixes for ibek-support
IBEK_GLOBALS = Path("_global")
SUPPORT_YAML_PATTERN = "*ibek.support.yaml"
PVI_YAML_PATTERN = "*pvi.device.yaml"

GLOBALS = _Globals()

JINJA = r".*\{\{.*\}\}.*"


class BaseSettings(BaseModel):
    """A Base class for setting consistent Pydantic model configuration"""

    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
    )


class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()



import os
import shutil
from pathlib import Path

from pytest import fixture
from pytest_mock import MockerFixture
from ruamel.yaml import YAML
from typer.testing import CliRunner

from ibek.__main__ import cli
from ibek.entity_factory import EntityFactory
from ibek.globals import GLOBALS
from ibek.support import Support

runner = CliRunner()


def run_cli(*args):
    result = runner.invoke(cli, [str(x) for x in args])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0, result


def get_support(yaml_file: str) -> Support:
    """
    Get a support object from the sample YAML directory
    """
    # load from file
    d = YAML(typ="safe").load(yaml_file)
    # create a support object from that dict
    support = Support(**d)
    return support


@fixture
def templates():
    return Path(__file__).parent.parent / "src" / "ibek" / "templates"


@fixture
def samples():
    return Path(__file__).parent / "samples"


@fixture
def support_defs(samples):
    return samples / "support"


@fixture
def ioc_defs(samples):
    return samples / "iocs"


# only one EntityFactory instance for each test
@fixture(scope="function")
def entity_factory():
    return EntityFactory()


@fixture
def tmp_epics_root(samples: Path, tmp_path: Path, mocker: MockerFixture):
    # create an partially populated epics_root structure in a temporary folder
    epics = tmp_path / "epics"
    epics_source = samples / "epics"
    epics.mkdir()
    # a dummy venv for testing extract_assets
    Path.mkdir(tmp_path / "venv")

    # create the minimal structure under epics root
    files = epics_source.glob("*")
    for f in files:
        if f.is_dir():
            shutil.copytree(f, epics / f.name)
        else:
            shutil.copy(f, epics / f.name)
    Path.mkdir(epics / "opi", exist_ok=True)
    Path.mkdir(epics / "epics-base", exist_ok=True)
    Path.mkdir(epics / "generic-source" / "ioc" / "config", parents=True)
    (epics / "ioc").symlink_to(epics / "generic-source" / "ioc")
    Path.mkdir(epics / "ibek-defs", exist_ok=True)
    Path.mkdir(epics / "runtime", exist_ok=True)

    # patch the global EPICS_ROOT to point to this temporary folder
    mocker.patch.object(GLOBALS, "_EPICS_ROOT", epics)

    # this should not be needed - what gives?
    os.environ["IOC"] = "/epics/ioc"
    os.environ["RUNTIME_DIR"] = "/epics/runtime"

    return epics


@fixture
def asyn_classes(support_defs, entity_factory):
    asyn_support = get_support(support_defs / "asyn.ibek.support.yaml")

    # make entity subclasses for everything defined in it
    namespace = entity_factory._make_entity_models(asyn_support)

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace


@fixture
def motor_classes(support_defs, entity_factory):
    asyn_support = get_support(support_defs / "asyn.ibek.support.yaml")
    motor_support = get_support(support_defs / "motorSim.ibek.support.yaml")

    # make entity subclasses for everything defined in it
    namespace = entity_factory._make_entity_models(asyn_support)
    namespace.extend(entity_factory._make_entity_models(motor_support))

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace



import logging
import os
import shutil
import subprocess
from pathlib import Path
from typing import List

import typer

from ibek.globals import GLOBALS

log = logging.getLogger(__name__)


def move_file(src: Path, dest: Path, binary: List[str]):
    """
    Move a file / tree / symlink from src to dest, stripping symbols from
    binaries if they are in the binary list.
    """
    dest.parent.mkdir(exist_ok=True, parents=True)

    if src.is_symlink():
        # copy the symlink
        shutil.rmtree(dest, ignore_errors=True)
        dest = dest.parent
        typer.echo(f"Symlink {src} to {dest}")
        shutil.copy(src, dest, follow_symlinks=False)
    else:
        typer.echo(f"Moving {src} to {dest}")
        if subprocess.call(["bash", "-c", f"mv {src} {dest}"]) > 0:
            raise RuntimeError(f"Failed to move {src} to {dest}")
    if dest.name in binary:
        # strip the symbols from the binary
        cmd = f"strip $(find {dest} -type f) &> /dev/null"
        subprocess.call(["bash", "-c", cmd])


def extract_assets(
    destination: Path,
    source: Path,
    extras: List[Path],
    defaults: bool,
    dry_run: bool = False,
):
    """
    extract and copy runtime assets from a completed developer stage container
    """

    if GLOBALS.STATIC_BUILD:
        # static builds only need database and .proto files from support modules
        asset_matches = "db|*/protocol"
    else:
        # dynamically linked builds need binaries and protocol files
        asset_matches = "bin|db|lib|*/protocol"

    # chdir out of the folders we will move
    os.chdir(source)

    # a default set of assets that all IOCs will need at runtime
    if defaults:
        default_assets = [
            source / "support" / "configure",
            GLOBALS.PVI_DEFS,
            GLOBALS.IBEK_DEFS,
            GLOBALS.IOC_FOLDER,  # get the IOC folder symlink
            Path.readlink(
                GLOBALS.IOC_FOLDER
            ).parent,  # get contents of IOC folder and its source (parent)
            GLOBALS.EPICS_ROOT.parent / "venv",  # virtualenv is a peer to /epics
        ]
    else:
        default_assets = []

    # folder names with binary files in them
    binary = ["bin", "lib"]

    # get the list of additional files supplied by add_runtime_files
    if GLOBALS.RUNTIME_FILES.exists():
        with GLOBALS.RUNTIME_FILES.open("r") as f:
            runtime_files = [Path(line.strip()) for line in f.readlines()]
    else:
        runtime_files = []

    # move the default assets and extras in their entirety
    extra_files = default_assets + extras + runtime_files
    for asset in extra_files:
        src = source / asset
        if src.exists():
            dest_file = destination / asset.relative_to("/")
            if dry_run:
                typer.echo(f"Would move extra asset {src} to {dest_file} with {binary}")
            else:
                move_file(src, dest_file, binary)
        else:
            typer.echo(f"WARNING: runtime asset {src} missing")

    # identify EPICS modules as folders with binary output folders
    # and move only their output folders as specified by asset_matches
    binaries: List[Path] = []
    for find in binary:
        # only look two levels deep
        binaries.extend(source.glob(f"*/*/{find}"))
        binaries.extend(source.glob(f"*/{find}"))

    modules = [binary.parent for binary in binaries]

    destination.mkdir(exist_ok=True, parents=True)
    for module in modules:
        # make sure dest folder exists
        destination_module = destination / module.relative_to("/")

        # use globs to make a list of the things we want to copy
        asset_globs = [module.glob(match) for match in asset_matches.split("|")]
        assets: List[Path] = [
            asset for asset_glob in asset_globs for asset in asset_glob
        ]

        for asset in assets:
            src = module / asset
            if src.exists():
                dest_file = destination_module / asset.relative_to(module)
                if dry_run:
                    typer.echo(f"Would move {src} to {dest_file} with {binary}")
                else:
                    move_file(src, dest_file, binary)



"""
Support Class to represent a deserialized <MODULE_NAME>.ibek.support.yaml file.
"""

from __future__ import annotations

import json
from typing import Any, Sequence

from pydantic import Field

from .entity_model import EntityModel
from .globals import BaseSettings


class Support(BaseSettings):
    """
    Lists the EntityModels for a support module, this defines what Entities it supports
    """

    shared: Sequence[Any] = Field(
        description="A place to create any anchors required for repeating YAML",
        default=(),
    )

    module: str = Field(description="Support module name, normally the repo name")
    entity_models: Sequence[EntityModel] = Field(
        description="The Entity Models an IOC can create using this module"
    )

    @classmethod
    def get_schema(cls):
        return json.dumps(cls.model_json_schema(), indent=2)



import operator

from semantic_version import Version


def semver_compare(base: str, target: str) -> bool:
    """
    Compare two semVer strings similarly to pip's requirements specifier syntax
    """
    ops = {
        ">=": operator.ge,
        "<=": operator.le,
        "==": operator.eq,
        ">": operator.gt,
        "<": operator.lt,
    }
    for op in ops.keys():
        if op in target:
            in_op = op
            target = target.strip(op)
            break

    if ops[in_op](Version.coerce(base), Version.coerce(target)):
        return True
    else:
        return False



"""
A class for rendering a substitution file from multiple instantiations of
support module yaml files.
"""

import re
from dataclasses import dataclass
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple

from ibek.entity_model import Database
from ibek.ioc import Entity
from ibek.utils import UTILS


def str_to_bool(v):
    return v.lower() in ("yes", "true", "1")


class RenderDb:
    @dataclass
    class RenderDbTemplate:
        filename: str
        rows: List[List[str]]
        columns: List[int]

    def __init__(self, entities: Sequence[Entity]) -> None:
        self.entities = entities
        # a mapping from template file name to details of instances of that template
        self.render_templates: Dict[str, RenderDb.RenderDbTemplate] = {}

    def add_row(self, filename: str, params: Mapping[str, Any], entity: Entity) -> None:
        """
        Accumulate rows of arguments for each template file,
        Adding a new template file if it does not already exist.
        Convert all arguments to strings.
        """
        filename = UTILS.render(dict(entity), filename)

        if filename not in self.render_templates:
            # for new filenames create a new RenderDbTemplate entry
            headings = [str(i) for i in list(params.keys())]
            self.render_templates[filename] = RenderDb.RenderDbTemplate(
                filename=filename,
                rows=[headings],  # first row is the headings
                columns=[0] * len(params),
            )

        # add a new row of argument values, rendering any Jinja template fields
        row = list(UTILS.render_map(dict(entity), params).values())

        # save the new row
        self.render_templates[filename].rows.append(row)

    def parse_instances(self) -> None:
        """
        Gather the database template instantiations from all entities
        while validating the arguments
        """
        for entity in self.entities:
            databases = entity._model.databases

            # Not all entities instantiate database templates
            if entity.entity_enabled and databases is not None:
                for database in databases:
                    self.add_database(database, entity)

    def add_database(self, database: Database, entity: Entity) -> None:
        """Validate database and add row using entity as context.

        Args:
            database: Database to add row for
            entity: Entity to use as context for Jinja template expansion

        """
        if str_to_bool(UTILS.render(entity, database.enabled)):
            database.file = database.file.strip("\n")

            parameters = entity.__dict__.keys()
            # take each database parameter and expand it into the set of
            # entity parameters that it matches using regex
            expanded_database_entries: dict[str, str | None] = {}
            for arg, value in database.args.items():
                if "*" in arg or "?" in arg:
                    # this is a regex - match the parameters to the regex
                    for parameter in parameters:
                        if re.match(arg, parameter):
                            # TODO - should we perform regex subst on the value as well?
                            expanded_database_entries[parameter] = value
                else:
                    # simple db argument name
                    expanded_database_entries[arg] = value

            for arg, value in expanded_database_entries.items():
                if value is None:
                    if arg not in entity.__dict__ and arg not in UTILS.variables:
                        raise ValueError(
                            f"database arg '{arg}' in database template "
                            f"'{database.file}' not found in context"
                        )

            self.add_row(database.file, expanded_database_entries, entity)

    def add_extra_databases(self, databases: List[Tuple[Database, Entity]]) -> None:
        """Add databases that are not part of EntityModels

        Args:
            databases: Databases to add, each mapped against an Entity to use as context

        """
        for database, entity in databases:
            self.add_database(database, entity)

    def align_columns(self) -> None:
        """
        Make sure columns will line up for each template file, also
        provide escaping for spaces and quotes
        """

        # first calculate the column width for each template
        # including escaping spaces and quotes
        for template in self.render_templates.values():
            for n, row in enumerate(template.rows):
                for i, arg in enumerate(row):
                    row[i] = f'"{row[i]}"'
                    if i < len(template.columns) - 1:
                        row[i] += ", "
                    template.columns[i] = max(template.columns[i], len(row[i]))

        # now pad each column to the maximum width
        for template in self.render_templates.values():
            for row in template.rows:
                for i, arg in enumerate(row):
                    row[i] = arg.ljust(template.columns[i])

    def render_database(
        self, extra_databases: Optional[List[Tuple[Database, Entity]]] = None
    ) -> Dict[str, List[str]]:
        """Render a database substitution file.

        Args:
            extra_databases: Databases to add that are not included on an Entity

        """
        extra_databases = [] if extra_databases is None else extra_databases

        self.parse_instances()
        self.add_extra_databases(extra_databases)
        self.align_columns()

        results = {}

        for template in self.render_templates.values():
            results[template.filename] = ["".join(row) for row in template.rows]

        return results



"""
Classes to specify arguments to Entity Models
"""

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, Optional

from pydantic import Field
from typing_extensions import Annotated, Literal

from .globals import JINJA, BaseSettings

JinjaString = Annotated[
    str, Field(description="A Jinja2 template string", pattern=JINJA)
]


class DefineType(Enum):
    """The type of a value"""

    string = "str"
    float = "float"
    int = "int"
    bool = "bool"
    list = "list"

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class Define(BaseSettings):
    """A calculated value for an Entity Model"""

    description: str = Field(
        description="Description of what the value will be used for"
    )
    value: Any = Field(description="The contents of the value")
    type: DefineType | None = Field(
        description="The type of the value", default=DefineType.string
    )


class Param(BaseSettings):
    """Base class for all Argument Types"""

    type: str

    description: str = Field(
        description="Description of what the argument will be used for"
    )


class FloatParam(Param):
    """An argument with a float value"""

    type: Literal["float"] = "float"
    default: Optional[float | JinjaString] = None


class StrParam(Param):
    """An argument with a str value"""

    type: Literal["str"] = "str"
    default: Optional[str] = None


class IntParam(Param):
    """An argument with an int value"""

    type: Literal["int"] = "int"
    default: Optional[int | JinjaString] = None


class BoolParam(Param):
    """An argument with an bool value"""

    type: Literal["bool"] = "bool"
    default: Optional[bool | JinjaString] = None


class ObjectParam(Param):
    """A reference to another entity defined in this IOC"""

    type: Literal["object"] = "object"
    # represented by an id string in YAML but converted to an Entity object
    # during validation
    default: Optional[str | object] = None


class IdParam(Param):
    """Explicit ID argument that an object can refer to"""

    type: Literal["id"] = "id"
    default: Optional[str] = None


class EnumParam(Param):
    """An argument with an enum value"""

    type: Literal["enum"] = "enum"
    default: Optional[Any] = None

    values: Dict[str, Any] = Field(
        description="provides a list of values to make this argument an Enum",
    )



"""
Functions for inserting snippets into files in an idempotent fashion
"""

from enum import Enum
from pathlib import Path
from typing import List

import typer

from ibek.globals import GLOBALS


class Local(str, Enum):
    non_local = ""
    local = ".local"


class Arch(str, Enum):
    none = ""
    common = ".Common"
    x86_64 = ".linux-x86_64"
    arm64 = ".linux-arm64"
    vxWorks = ".vxWorks-ppc604_long"
    rtems = ".RTEMS-beatnik"


def get_config_site_file(
    module: str,
    host: Arch = Arch.x86_64,
    target: Arch = Arch.common,
) -> Path:
    """
    Return the path to an CONFIG_SITE file for a support module.
    Default is /epics/module/configure/CONFIG_SITE.local.linux-x86_64.Common
    """
    name = f"CONFIG_SITE{host.value}{target.value}"
    filepath = GLOBALS.SUPPORT / module / "configure" / name

    return filepath


def get_release_file(
    module: str,
    host: Arch = Arch.none,
    target: Arch = Arch.none,
    local: Local = Local.local,
) -> Path:
    """
    Return the path to an EPICS RELEASE file for a support
    default is /epics/module/configure/RELEASE.local
    """
    name = f"RELEASE{host.value}{target.value}{local.value}"
    filepath = GLOBALS.SUPPORT / module / "configure" / name

    return filepath


def add_list_to_file(file: Path, text_list: List[str]):
    """
    add a sequence of strings into a file, leaving previously existing
    strings where they are
    """
    if len(text_list) == 0:
        return

    for line in text_list:
        # skip blanks and comments
        if line.startswith("#") or line == "":
            continue
        add_text_once(file, line)


def add_text_once(file: Path, text: str):
    """
    Idempotent add of text to a file.
    Creates the file if it doesn't exist.
    """

    if not file.exists():
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(text + "\n")
    else:
        current = file.read_text()
        if text not in current:
            file.write_text(current + text + "\n")


def symlink_files(source_directory: Path, file_pattern: str, target_directory: Path):
    """
    Symlink files matching the given pattern in source directory to target directory.

    Args:
        source_directory: Directory containing source files
        file_pattern: Pattern of files in source directory to be symlinked
        target_directory: Directory to create symlinks in

    """
    source_files = list(source_directory.glob(file_pattern))
    if not source_files:
        return

    typer.echo(f"Symlinking {file_pattern} files:")
    target_directory.mkdir(parents=True, exist_ok=True)
    for source_file in source_files:
        typer.echo(f" {target_directory / source_file.name} -> {source_file}")
        target = target_directory / source_file.name
        target.unlink(missing_ok=True)
        target.symlink_to(source_file)



"""
functions for making Entity models
"""

from __future__ import annotations

from pathlib import Path
from typing import Annotated, List, Sequence, Type, Union

from pydantic import Field
from ruamel.yaml.main import YAML

from ibek.utils import UTILS

from .ioc import IOC, Entity


class IocFactory:
    """
    A Class for creating Generic IOC classes and instances
    """

    def deserialize_ioc(
        self, ioc_instance_yaml: Path, entity_models: List[Type[Entity]]
    ) -> IOC:
        """
        Takes an ioc instance entities file, list of generic ioc yaml files.

        Returns a model of the resulting ioc instance
        """

        # Create a Class representing this Generic IOC
        ioc_model = self.make_ioc_model(entity_models)

        # extract the ioc instance yaml into a dict
        ioc_instance_dict = YAML(typ="safe").load(ioc_instance_yaml)

        if ioc_instance_dict is None or "ioc_name" not in ioc_instance_dict:
            raise RuntimeError(
                f"Failed to load a valid ioc config from {ioc_instance_yaml}"
            )

        # extract the ioc name into UTILS for use in jinja renders
        name = UTILS.render({}, ioc_instance_dict["ioc_name"])
        UTILS.set_ioc_name(name)
        ioc_instance_dict["ioc_name"] = name

        # Create an IOC instance from the instance dict and the model
        ioc_instance = ioc_model(**ioc_instance_dict)

        return ioc_instance

    def make_ioc_model(self, entity_models: List[Type[Entity]]) -> Type[IOC]:
        """
        Create an IOC derived model, by setting its entities attribute to
        be of type 'list of Entity derived classes'.

        Also instantiate any SubEntities at this time
        """

        entity_union = Union[tuple(entity_models)]  # type: ignore
        discriminated = Annotated[entity_union, Field(discriminator="type")]  # type: ignore

        class NewIOC(IOC):
            entities: Sequence[discriminated] = Field(  # type: ignore
                description="List of entities this IOC instantiates"
            )

        return NewIOC



"""
A class containing utility functions for passing into the Jinja context.

This allows us to provide simple functions that can be called inside
Jinja templates with {{ _global.function_name() }}. It also allows
us to maintain state between calls to the Jinja templates because
we pass a single instance of this class into all Jinja contexts.
"""

import os
from pathlib import Path
from typing import Any, Dict, Mapping

from jinja2 import StrictUndefined, Template


class Utils:
    """
    A Utility class for adding functions to the Jinja context
    """

    def __init__(self: "Utils"):
        self.file_name: str = ""
        self.ioc_name: str = ""
        self.__reset__()

    def __reset__(self: "Utils"):
        """
        Reset all saved state. For use in testing where more than one
        IOC is rendered in a single session
        """
        self.variables: Dict[str, Any] = {}

    def set_file_name(self: "Utils", file: Path):
        """
        Set the ioc name based on the file name of the instance definition
        """
        self.file_name = file.stem

    def set_ioc_name(self: "Utils", name: str):
        """
        Set the ioc name based on the file name of the instance definition
        """
        self.ioc_name = name

    def get_env(self, key: str) -> str:
        """
        Get an environment variable
        """
        return os.environ.get(key, "")

    def set(self, key: str, value: Any) -> Any:
        """create a global variable for our jinja context"""
        s_key = str(key)
        self.variables[s_key] = value
        return value

    def get(self, key: str, default="") -> Any:
        """get the value a global variable for our jinja context"""
        # default is used to set an initial value if the variable is not set
        s_key = str(key)
        return self.variables.get(s_key, default)

    def incrementor(
        self, name: str, start: int = 0, increment: int = 1, stop: int | None = None
    ) -> int:
        """
        get a named counter that increments by inc each time it is called

        creates a new counter if it does not yet exist
        """
        index = str(name)
        counter = self.variables.get(index)

        if counter is None:
            self.variables[index] = start
        else:
            if not isinstance(counter, int):
                raise ValueError(f"Variable {index} is not an integer")
            self.variables[index] += increment
            if stop is not None and self.variables[index] > stop:
                raise ValueError(f"Counter {index} exceeded maximum value of {stop}")

        return self.variables[index]

    def render(self, context: Any, template_text: Any) -> str:
        """
        Render a Jinja template with the global _global object in the context
        """
        if not isinstance(template_text, str):
            # because this function is used to template arguments, it may
            # be passed a non string which will always render to itself
            return template_text

        try:
            jinja_template = Template(template_text, undefined=StrictUndefined)
            return jinja_template.render(
                context,
                # global context for all jinja renders
                _global=self,
                # put variables created with set/get directly in the context
                **self.variables,
                ioc_yaml_file_name=self.file_name,
                ioc_name=self.ioc_name,
            )
        except Exception:
            print(f"ERROR RENDERING TEMPLATE:\n{template_text}")
            raise

    def render_map(self, context: Any, map: Mapping[str, str | None]) -> dict[str, str]:
        """
        Render a map of jinja templates with values from the given context.

        If given a key with a value of `None`, the key itself will be used as a template
        value, so ``{"P": None}`` is equivalent to ``{"P": "{{ P }}"}``.

        Args:
            context: Context to extract template variables from
            map: Map of macro to jinja template to render

        """
        return {
            key: self.render(
                context, template if template is not None else "{{ %s }}" % key
            )
            for key, template in map.items()
        }


# a singleton Utility object for sharing state across all Entity renders
UTILS: Utils = Utils()



from pydantic import BaseModel, Field


class SubEntity(BaseModel, extra="allow"):
    """
    A loosely defined class to declare the Entities
    in an ibek.support.yaml file in the 'sub_entities' property of an Entity
    section
    """

    type: str = Field(description="The type of this entity")

    entity_enabled: bool = Field(
        description="enable or disable this entity instance", default=True
    )



import shutil
from pathlib import Path
from typing import List, Tuple

import typer
from pvi._format.base import IndexEntry
from pvi._format.dls import DLSFormatter
from pvi._format.template import format_template
from pvi.device import Device

from ibek.entity_factory import EntityFactory
from ibek.entity_model import Database
from ibek.gen_scripts import create_boot_script, create_db_script
from ibek.globals import GLOBALS, NaturalOrderGroup
from ibek.ioc import IOC, Entity
from ibek.ioc_factory import IocFactory
from ibek.utils import UTILS

runtime_cli = typer.Typer(cls=NaturalOrderGroup)


@runtime_cli.command()
def generate(
    instance: Path = typer.Argument(
        ...,
        help="The filepath to the ioc instance entity file",
        autocompletion=lambda: [],  # Forces path autocompletion
    ),
    definitions: List[Path] = typer.Argument(
        ...,
        help="The filepath to a support module yaml file",
        autocompletion=lambda: [],  # Forces path autocompletion
    ),
):
    """
    Build a startup script for an IOC instance
    """
    # the file name under of the instance definition provides the IOC name
    UTILS.set_file_name(instance)

    entity_factory = EntityFactory()
    entity_models = entity_factory.make_entity_models(definitions)
    ioc_instance = IocFactory().deserialize_ioc(instance, entity_models)

    # post processing to insert SubEntity instances
    all_entities = entity_factory.resolve_sub_entities(ioc_instance.entities)
    ioc_instance.entities = all_entities

    # Clear out generated files so developers know if something stops being generated
    shutil.rmtree(GLOBALS.RUNTIME_OUTPUT, ignore_errors=True)
    GLOBALS.RUNTIME_OUTPUT.mkdir(exist_ok=True)
    shutil.rmtree(GLOBALS.OPI_OUTPUT, ignore_errors=True)
    GLOBALS.OPI_OUTPUT.mkdir(exist_ok=True)

    pvi_index_entries, pvi_databases = generate_pvi(ioc_instance)
    generate_index(ioc_instance.ioc_name, pvi_index_entries)

    script_txt = create_boot_script(ioc_instance.entities)
    script_output = GLOBALS.RUNTIME_OUTPUT / "st.cmd"
    script_output.parent.mkdir(parents=True, exist_ok=True)
    with script_output.open("w") as stream:
        stream.write(script_txt)

    db_txt = create_db_script(ioc_instance.entities, pvi_databases)
    db_output = GLOBALS.RUNTIME_OUTPUT / "ioc.subst"
    with db_output.open("w") as stream:
        stream.write(db_txt)


def generate_pvi(ioc: IOC) -> Tuple[List[IndexEntry], List[Tuple[Database, Entity]]]:
    """Generate pvi bob and template files to add to UI index and IOC database.

    Args:
        ioc: IOC instance to extract entity pvi definitions from

    Returns:
        List of bob files to add as buttons on index and databases to add to IOC
        substitution file

    """
    index_entries: List[IndexEntry] = []
    databases: List[Tuple[Database, Entity]] = []

    formatter = DLSFormatter()

    formatted_pvi_devices: List[str] = []
    for entity in ioc.entities:
        definition = entity._model
        if not hasattr(definition, "pvi") or definition.pvi is None:
            continue
        entity_pvi = definition.pvi

        pvi_yaml = GLOBALS.PVI_DEFS / entity_pvi.yaml_path
        device_name = pvi_yaml.name.split(".")[0]
        device_bob = GLOBALS.OPI_OUTPUT / f"{device_name}.pvi.bob"

        # Skip deserializing yaml if not needed
        if entity_pvi.pv or device_name not in formatted_pvi_devices:
            device = Device.deserialize(pvi_yaml)
            device.deserialize_parents([GLOBALS.PVI_DEFS])

            if entity_pvi.pv:
                # Create a template with the V4 structure defining a PVI interface
                output_template = GLOBALS.RUNTIME_OUTPUT / f"{device_name}.pvi.template"
                format_template(device, entity_pvi.pv_prefix, output_template)

                # Add to extra databases to be added into substitution file
                databases.append(
                    (
                        Database(file=output_template.name, args=entity_pvi.ui_macros),
                        entity,
                    )
                )

            if device_name not in formatted_pvi_devices:
                formatter.format(device, device_bob)

                # Don't format further instance of this device
                formatted_pvi_devices.append(device_name)

        if entity_pvi.ui_index:
            macros = UTILS.render_map(dict(entity), entity_pvi.ui_macros)
            index_entries.append(
                IndexEntry(
                    label=f"{device.label}",
                    ui=device_bob.name,
                    macros=macros,
                )
            )

    return index_entries, databases


def generate_index(title: str, index_entries: List[IndexEntry]):
    """Generate an index bob using pvi.

    Args:
        title: Title of index UI
        index_entries: List of entries to include as buttons on index UI

    """
    DLSFormatter().format_index(title, index_entries, GLOBALS.OPI_OUTPUT / "index.bob")



"""
System tests that run the CLI commands and compare the output to expected
results.
"""

import json
import subprocess
import sys
from pathlib import Path

from pytest_mock import MockerFixture

import ibek.utils as utils
from ibek import __version__
from ibek.globals import (
    GLOBALS,
    PVI_YAML_PATTERN,
    SUPPORT_YAML_PATTERN,
)
from ibek.runtime_cmds.commands import generate
from ibek.support_cmds.commands import generate_links
from tests.conftest import run_cli


def test_cli_version():
    cmd = [sys.executable, "-m", "ibek", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def test_ibek_schema(tmp_path: Path, samples: Path):
    """generate the global ibek schema"""
    schema_path = tmp_path / "schema.json"
    run_cli("support", "generate-schema", "--output", schema_path)
    expected = json.loads(
        (samples / "schemas" / "ibek.support.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_single_schema(tmp_path: Path, samples: Path):
    """generate schema from a support module definition yaml"""

    schema_path = tmp_path / "single.ibek.support.schema.json"
    yaml_path = samples / "support" / "motorSim.ibek.support.yaml"
    run_cli(
        "ioc", "generate-schema", "--no-ibek-defs", yaml_path, "--output", schema_path
    )

    expected = json.loads(
        (samples / "schemas" / "single.ibek.ioc.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_motor_sim_schema(tmp_path: Path, samples: Path):
    """generate schema for a container with two support modules"""

    schema_combined = tmp_path / "motorSim.ibek.ioc.schema.json"
    yaml_path1 = samples / "support" / "asyn.ibek.support.yaml"
    yaml_path2 = samples / "support" / "motorSim.ibek.support.yaml"
    run_cli(
        "ioc",
        "generate-schema",
        "--no-ibek-defs",
        yaml_path1,
        yaml_path2,
        "--output",
        schema_combined,
    )

    expected = json.loads(
        (samples / "schemas" / "motorSim.ibek.ioc.schema.json").read_text()
    )

    actual = json.loads((schema_combined).read_text())
    assert expected == actual


def test_build_runtime_motorSim(tmp_epics_root: Path, samples: Path):
    """
    build an ioc runtime script from an IOC instance entity file
    and multiple support module definition files

    Also verifies database subst file generation for multiple
    entity instantiations.
    """

    generic_generate(
        tmp_epics_root,
        samples,
        "motorSim",
        ["motorSim", "asyn"],
    )


def test_build_utils_features(tmp_epics_root: Path, samples: Path):
    """
    build an ioc runtime script to verify utils features
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "utils",
        ["utils"],
    )


def test_generate_links_ibek(samples: Path, mocker: MockerFixture):
    symlink_mock = mocker.patch("ibek.support_cmds.commands.symlink_files")

    generate_links(samples / "support")

    symlink_mock.assert_any_call(
        samples / "support", PVI_YAML_PATTERN, GLOBALS.PVI_DEFS
    )
    symlink_mock.assert_any_call(
        samples / "support", SUPPORT_YAML_PATTERN, GLOBALS.IBEK_DEFS
    )


def test_ipac(tmp_epics_root: Path, samples: Path):
    """
    Tests that an id argument can include another argument in its default value
    """

    # reset the InterruptVector counter to its initial state (if already used)
    if "InterruptVector" in utils.UTILS.variables:
        utils.UTILS.variables["InterruptVector"] = 191

    generic_generate(
        tmp_epics_root,
        samples,
        "ipac-test",
        ["ipac", "epics"],
    )


def test_gauges(tmp_epics_root: Path, samples: Path):
    """
    Tests that an id argument can include another argument in its default value
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "gauges",
        ["asyn", "gauges"],
    )


def test_quadem(tmp_epics_root: Path, samples: Path):
    """
    Tests the use of CollectionDefinitions in an IOC instance
    this example uses the tetramm beam position monitor module
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "quadem",
        ["ADCore", "quadem"],
    )


def generic_generate(
    epics_root: Path, samples: Path, ioc_name: str, support_names: list[str]
):
    ioc_yaml = samples / "iocs" / f"{ioc_name}.ibek.ioc.yaml"
    support_yamls = [
        samples / "support" / f"{name}.ibek.support.yaml" for name in support_names
    ]
    expected_outputs = samples / "outputs" / ioc_name

    generate(ioc_yaml, support_yamls)

    for output in expected_outputs.glob("*"):
        actual = epics_root / "runtime" / output.name
        if not actual.exists():
            actual = epics_root / "opi" / output.name
        assert actual.exists(), "Missing output file"
        assert output.read_text() == actual.read_text()


def test_andreas_motors(tmp_epics_root: Path, samples: Path):
    """
    Motor and axis example
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "technosoft",
        ["asyn", "technosoft"],
    )


def test_list(tmp_epics_root: Path, samples: Path):
    """
    Motor and axis example
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "listarg",
        ["listarg"],
    )


def test_fast_vacuum(tmp_epics_root: Path, samples: Path):
    """
    Cut down copy of dlsPLC containing fast vacuum master and channel
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "fastVacuum",
        ["fastVacuum"],
    )


def test_dls_plc(tmp_epics_root: Path, samples: Path):
    """
    Cut down copy of dlsPLC containing fast vacuum master and channel
    """
    generic_generate(
        tmp_epics_root,
        samples,
        "dlsPLC",
        ["dlsPLC", "asyn"],
    )



#!/usr/bin/env python
"""
Converts .ibek.support.yaml files from the v2.0 format to the v3.0 format.
Can also convert ioc.yaml files as the changes re __utils__ are the same.

The schema changes are as follows:

- `defs` are now `entity_models`
- `args` are now `parameters`
- `parameters` are a dictionary of dictionaries instead of a list of dictionaries
- `parameters` no longer have a name field as the key in the dictionary is the name
- `values` is now `post_defines` and is also a dictionary of dictionaries
- `pre_values` is now `pre_defines` and is also a dictionary of dictionaries
- `__utils__` is now `_global`
- `_global` functions have changed:
    - `get_var` -> `get`
    - `set_var` -> `set`
    - `counter` -> `incrementor`
      - behaviour has changed: the --increment and --stop flags are interpreted
        on every invocation of the incrementor function (so non linear counters
        are possible)
      - incrementors current value is accessible via _global.get(incrementorName)

Backward compatible changes:
- object refs are literally the object - so you can modify a referenced object's
  attributes (with appropriate care!!)
- `pre_defines` and `post_defines` now may have optional `type:` which defaults to `str` and may be
    - `list` (of Any), `int`, `float`, `bool`
- each `entity_model` may have a `shared:` field which is a `list` of `Any`
- there is also a global `shared:` at the root level
- `shared:` are scratch areas in which to place anchors that can be referred to by aliases - and are therefore useful for placing repeating patterns (especially of 'parameters') and then using aliases where these repeats are required.

"""

import re
from pathlib import Path

import typer
from ruamel.yaml import YAML, CommentedMap


class ConvertedAlready(Exception):
    pass


def main(files: list[Path]):
    """
    Read a list of files in and process each one
    """
    for f in files:
        print(f"Processing {f} ...")
        process_file(f)


def tidy_up(yaml):
    # add blank lines between major fields - this is an opinionated
    # formatting choice for legibility
    for field in [
        "  - name:",
        "    databases:",
        "    pre_init:",
        "    post_init:",
        "    pre_defines:",
        "    post_defines:",
        "    shared:",
        "module",
        "entity_models",
        "      - file:",
        "      - value:",
    ]:
        # insert a blank line before the field unless it already has one
        yaml = re.sub(r"^[ \t]*(\n%s)" % field, "\n\\g<1>", yaml)
        yaml = re.sub(r"__utils__", "_global", yaml)
        yaml = re.sub(r"_ctx_", "_global", yaml)
        yaml = re.sub(r"\.counter", ".incrementor", yaml)
        yaml = re.sub(r"\.get_var", ".get", yaml)
        yaml = re.sub(r"\.setvar", ".incrementor", yaml)
    return yaml


def report(message: str):
    print(f">>>>>>> {message}")


def process_file(file: Path):
    """
    process a single file by reading its yaml - transforming it and writing it back
    """
    yaml = YAML()
    with open(file, "r") as f:
        data = yaml.load(f)

    try:
        if data is None:
            report("empty file !!")
            return

        if "module" in data:
            report("module yaml")
            convert(data)
        elif "ioc_name" in data:
            report("ioc yaml")
        else:
            report("not a supported yaml file !!")
            return

    except ConvertedAlready:
        report("already converted")
    else:
        with open(file, "w") as f:
            yaml.default_flow_style = False
            # these indents match the defaults for vscode RedHat yaml extension
            yaml.indent(mapping=2, sequence=4, offset=2)
            # we like to allow really wide lines for db subst lists etc so
            # we recommend this setting for the yaml extension
            yaml.width = 8000
            yaml.preserve_quotes = True
            yaml.dump(data, f, transform=tidy_up)


def convert(data: dict):
    """
    root data conversion function, replaces each definition in the data with
    a processed version of that definition
    """

    # replace all definitions with converted versions
    if "defs" in data:
        data["entity_models"] = [
            convert_definition(definition) for definition in data["defs"]
        ]
        del data["defs"]


def convert_definition(data: dict) -> dict | None:
    """
    convert a single definition's args list to a parameters dict
    """

    # rebuild the definition from scratch to enforce order
    new_data = CommentedMap()

    # copy the leading keys that are not being changed
    copy_verbatim(data, new_data, ["name", "description"])

    try:
        if "pre_defines" in data:
            check_converted(data["pre_defines"])
            new_data["pre_defines"] = list_to_dict(data["pre_defines"])
        if "args" in data:
            new_data["parameters"] = list_to_dict(data["args"])
        else:
            raise ConvertedAlready  # probably! (or its not a support yaml file)
        if "post_defines" in data:
            check_converted(data["post_defines"])
            new_data["post_defines"] = list_to_dict(data["post_defines"])
    except ConvertedAlready:
        raise
    except Exception:
        print(f"Failed to convert {data}")
        raise

    # copy the trailing keys that are not being changed
    copy_verbatim(
        data,
        new_data,
        [
            "pre_init",
            "post_init",
            "databases",
            "env_vars",
            "pvi",
            "shared",
            "sub_entities",
        ],
    )

    return new_data


def check_converted(item: dict | list):
    if isinstance(item, dict):
        raise ConvertedAlready


def copy_verbatim(source: dict, dest: dict, keys: list[str]):
    """
    copy keys from data to a new dictionary
    """
    for key in keys:
        if key in source:
            dest[key] = source[key]


def list_to_dict(args: list[dict]) -> dict[str, dict]:
    """
    convert a list of args to a dictionary, removing the name field as that
    becomes the key in the dictionary
    """
    parameters = CommentedMap()
    try:
        for arg in args:
            name = arg["name"]
            parameters[name] = arg.copy()
            # name is the param key so we no longer need it
            del parameters[name]["name"]
        return parameters
    except Exception:
        print(f"Failed to convert {arg}")
        raise


if __name__ == "__main__":
    typer.run(main)
